import os
import sys
import importlib.util

import lift.print_color as Out
import lift.helpers as helpers
from lift.lift_class import LiftClass
import lift.template_lift_build as Template

def main():
    Out.print_info("- - - lift - - -")

    # CLI parsing
    if len(sys.argv) == 1:
        Out.print_help()
        exit()

    first_argument = sys.argv[1]
    second_argument = None
    if len(sys.argv) >= 3:
        second_argument = sys.argv[2]

    if first_argument == "init":
        Out.print_debug("> init")
        if not os.path.exists(os.getcwd() + "/lift_build.py"):
            with open(os.getcwd() + "/lift_build.py", "w") as file:
                file.write(Template.template)
                Out.print_info("> lift_build.py is created form a template")
        else:
            Out.print_error("> lift_build.py already exist")

    # Working dir which lift was called from, must contain lift_build.py
    lift_build_path = helpers.get_lift_build_path()
    if not os.path.isfile(lift_build_path):
        Out.print_error("lift_build.py is not found")
        exit()

    # Load the module using import_module
    spec = importlib.util.spec_from_file_location("lift_build", lift_build_path)
    lift_build_py = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(lift_build_py)

    # Passing default LiftClass to users lift_build.py->setup(lift) to conifigure
    build_LiftClass = LiftClass()
    Out.print_debug("> lift_build.py->setup(lift)")
    build_LiftClass = lift_build_py.setup(build_LiftClass)


    if first_argument == "build" or first_argument == "run":
        # build mode selection
        if second_argument == "release":
            Out.print_debug("> lift_build.py->build(\"release\")")
            lift_build_py.build(build_LiftClass, "release")
        else:
            Out.print_debug("> lift_build.py->build(\"debug\")")
            lift_build_py.build(build_LiftClass, "debug")
        # run the app's executable
        if first_argument == "run":
            Out.print_debug("> lift_build.py->run()")
            lift_build_py.run(build_LiftClass)
    elif first_argument == "test":
        Out.print_debug("> lift_build.py->test()")
        lift_build_py.test(build_LiftClass)
    elif first_argument == "clean":
        Out.print_debug("> lift_build.py->clean()")
        lift_build_py.clean(build_LiftClass)

if __name__ == "__main__":
    main()



"""
Entrypoint for incremental compilation.

Internally it uses FileModifiedCache and DependencyGraph to resolve a source directory
into a set of .c files that must be compiled into object files
"""
import os
import re
import lift.print_color as Out
from lift.file_modified_cache import FileModifiedCache
from lift.dependency_graph import DependencyGraph, FileNode
from lift.files import Files

class Incremental:
    def __init__(self, VERBOSE=True):
        self.VERBOSE = VERBOSE

    def resolve(self, path_src_files, build_dir):

        # exclude files which are not *.h and *.c 
        header_files = path_src_files.get_files_with_extensions({".h"})
        source_files = path_src_files.get_files_with_extensions({".c"})
        path_src_files = header_files + source_files

        cache = FileModifiedCache()
        cache.load()
        modified_files = []
        for file in path_src_files:
            mtime = os.path.getmtime(file)
            if cache.mtime_cache.get(file):
                if mtime > float(cache.mtime_cache[file]):
                    if self.VERBOSE:
                        Out.print_color(Out.COLOR.BLUE, f'{file} has been modified since last compilation')
                    modified_files.append(file)

            cache.add_file(file) # update cache
        
        cache.store() # at end of compilation we should store back to disk
        
        ### Build Dependency graph
        dgraph = DependencyGraph()
        file_nodes = [ FileNode(p, os.path.basename(p)) for p in path_src_files]
        dgraph.add_nodes(file_nodes)

        for file in file_nodes:
            # parse file for includes
            f = open(file.absolute_path, "r")
            file_contents = f.read()
            includes = extract_includes(file_contents)
            resolve_header_path_for_file = lambda include: resolve_relative_header_path(file.absolute_path, include)
            absolute_path_includes = map(resolve_header_path_for_file, includes)
            for include in absolute_path_includes:
                target_node = dgraph.get_node_by_absolute_path(include)
                if target_node:
                    dgraph.add_dependency(file, target_node, directional=True)
                else:
                    # Out.print_debug(f"couldnt find {include}")
                    pass


        incremental_compile_files = set()

        for file in modified_files:
            if file.endswith('.c'):
                incremental_compile_files.add(file)
            else: # header file
                start_node = dgraph.get_node_by_absolute_path(file)
                if start_node:
                    deps = dgraph.accumulate_dependencies(start_node)
                    for dep in deps:
                        incremental_compile_files.add(dep)
                else:
                    #TODO (Josh) how to handle error here?
                    pass

        # get object files that exist
        build_files = Files(build_dir).get_files_with_extensions({".o"})
        object_files = set([os.path.splitext(os.path.basename(path))[0] for path in build_files])

        c_files = source_files
        # we need to turn them into a dict where { [key]: value } key = 'magic' and value = 'full_path/src/magic.c'
        # so that we can compare them without their paths or their extension (just the 'magic'), but after the comparison
        # we want to retain the original full path of the source file to pass to the compiler.
        # to do this we turn both into dicts, do a comparison on the keys (same as set minus operation)
        # then use those keys to get back out a set of paths via indexing into the original dict
        c_files_name_to_path_map = {os.path.splitext(os.path.basename(path))[0]: path for path in c_files}
        o_files_name_to_path_map = {os.path.splitext(os.path.basename(path))[0]: path for path in object_files}
        #files *without* a .o is the set of .c files minus the set of .o files
        difference = c_files_name_to_path_map.keys() - o_files_name_to_path_map.keys()
        uncompiled_source_files = {c_files_name_to_path_map[key] for key in difference}

        # Files to compile are set of incremental compilation targets, I,  and uncompiled source files, S (I union S)
        source_files_to_compile = uncompiled_source_files | incremental_compile_files
        
        return source_files_to_compile

def resolve_relative_header_path(src_file_abs_path, include):
    src_dir = os.path.dirname(src_file_abs_path)
    header_abs_path = os.path.abspath(os.path.join(src_dir, include))
    return header_abs_path

def extract_includes(text):
    # the regular expression pattern to match "#include"
    pattern = re.compile(r'#include\s+[<"](.*?)[>"]')

    # find all matches in the input text
    matches = re.findall(pattern, text)

    # return the list of matches
    return matches



import lift.print_color as Out

# Flags for GCC / Clang
class Compiler:
    def __init__(self, which_compiler):
        self.CURRENT_COMPILER = which_compiler

    GCC = "gcc"
    CLANG = "clang"
    ALL = "all"
    GCC_COLORS="export GCC_COLORS='error=01;31:warning=01;35:note=01;36:caret=01;32:locus=01:quote=01'"
    RELEASE = "release"
    DEBUG = "debug"
    HARDENED = "hardened"

    CURRENT_COMPILER = CLANG

    flags_debug = {
        "-g": ALL,
        "-Wall": ALL,
        "-fsanitize=address,undefined": ALL,
        "-Wall": ALL,
        "-Wextra": ALL,
        "-fno-omit-frame-pointer": ALL,
        "-Wshadow": ALL,
        "-Wnull-dereference": ALL,
        "-Wcast-qual": ALL,
        "-Wformat-security": ALL,
        "-Wstack-protector": ALL,
        "-Warray-bounds": ALL,

        "-Wcast-align=strict": GCC,
        "-Wcast-align": CLANG,
        "-fcolor-diagnostics": CLANG,
        "-fdiagnostics-color=always": GCC,
        "-Warray-bounds-pointer-arithmetic": CLANG,
        "-Wassign-enum": CLANG,
    }

    flags_hardened = {
        "-fstack-protector": ALL,
    }

    flags_release = {
        "-O3": ALL,
    }

    def generate_flags(self, compilation_mode):
        flags = ""
        if compilation_mode == self.DEBUG:
            for flag in self.flags_debug:
                if self.flags_debug[flag] == self.ALL or self.flags_debug[flag] == self.CURRENT_COMPILER:
                    flags += f" {flag}"
        elif compilation_mode == self.RELEASE:
            for flag in self.flags_release:
                if self.flags_release[flag] == self.ALL or self.flags_release[flag] == self.CURRENT_COMPILER:
                    flags += f" {flag} "
        elif compilation_mode == self.HARDENED:
            for flag in self.flags_hardened:
                if self.flags_hardened[flag] == self.ALL or self.flags_hardened[flag] == self.CURRENT_COMPILER:
                    flags += f" {flag}"
        else:
            Out.print_error(f"{compilation_mode} is not a valid compilation mode")
        return flags;



# Macros for printing with colored tags

TOGGLE_DEBUG = True
 
def print_normal(text):
    print(f"{COLOR.RESET}{text}")
    
def print_info(text):
    print(f"{STYLE.DEFAULT}{text}{COLOR.RESET}")

def print_error(text):
    print(f"{STYLE.ERROR_TAG}ERROR:{COLOR.RESET} {text}")

def print_debug(text):
    if TOGGLE_DEBUG:
        print(f"{STYLE.DEBUG_TAG}DEBUG:{COLOR.RESET} {text}")

def print_color(color, text):
    print(f"{color}{text}{COLOR.RESET}")

def print_help():
    print_info("> HELP PAGE")
    print_normal("FIXME: HELP PAGE is not implemented yet")
# used for colored printing, example of use
# print(f"{COLOR.BLUE}{COLOR.BG.BLACK}blue on black{COLOR.RESET} normal")
class COLOR:
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    RESET = "\033[39;49m"

    class BG:
        BLACK = "\033[40m"
        RED = "\033[41m"
        GREEN = "\033[42m"
        YELLOW = "\033[43m"
        BLUE = "\033[44m"
        MAGENTA = "\033[45m"
        CYAN = "\033[46m"
        WHITE = "\033[47m"
        RESET = "\033[39;49m"


class STYLE:
    DEFAULT = f"{COLOR.YELLOW}"
    DEBUG = f"{COLOR.BLUE}"
    DEBUG_TAG = f"{COLOR.BLACK}{COLOR.BG.YELLOW}"
    ERROR = f"{COLOR.RED}{COLOR.BG.BLACK}"
    ERROR_TAG = f"{COLOR.BLACK}{COLOR.BG.RED}"





"""
Incremental compilation requires the tracking of modified source files.
When a file has been changed, we can use this to compute what recompilation is required.

A cache is comprised of a relative path from the root project directory, and a last modified timestamp
stored in a comma seperated list, with one file per line.

E.g.

<filepath>, timestamp
"""

import os
import pathlib
import re
import lift.helpers as helpers
import lift.print_color as Out

class FileModifiedCache:
    parse_re = re.compile(r'"(.*)",(.*)')

    def __init__(self):
        self.mtime_cache = {}       # lookup table of file -> mtime

    def get_cached_files(self):
        return self.mtime_cache.keys()

    """Persist self.cached_files to disk"""
    def store(self):
        cache_dir = helpers.build_dir_path() + ".cache/"

        # make directory if not already exists
        pathlib.Path(cache_dir).mkdir(parents=True, exist_ok=True) 

        with open(cache_dir + "file_mtimes", "w") as cache_file:
            for filepath, file_mtime in self.mtime_cache.items():
                # save to <project_root>/build/.cache/file_mtimes
                cache_file.write(f'"{filepath}",{file_mtime}\n')

    def load(self):
        cache_dir = helpers.build_dir_path() + ".cache/"
        # load from <project_root>/build/.cache/file_mtimes
        try:
            with open(cache_dir + "file_mtimes", "r") as cache_file:
                line = cache_file.readline()
                while line:
                    filepath, mtime = self._parse_line(line)
                    if not (filepath and mtime):
                        Out.print_error("Error parsing mtime file cache. Try running lift clean")
                        exit(1)

                    self.mtime_cache[filepath] = mtime

                    line = cache_file.readline()
        except FileNotFoundError as e:
            # do nothing. cache is empty
            pass

    def add_file(self, filepath):
        mtime = os.path.getmtime(filepath)
        self.mtime_cache[filepath] = mtime

    def clear(self):
        self.mtime_cache = {}

    def _parse_line(self, line):
        match = self.parse_re.search(line)
        if match:
            return match.group(1), match.group(2)
        
        return None, None


import os

# Use example:
# files = FILES("sample")
# lift_print(files.all_files)
# lift_print(files.get_files_with_extensions({".h",".c"}));
class Files:
    all_files = []
    def __init__(self, directory):
        if not os.path.exists(directory):
            os.mkdir(directory)
        self.all_files = self.internal_get_all_files(directory)
        
    def internal_get_all_files(self, directory):
        file_paths = []
        for root, directories, files in os.walk(directory):
            for filename in files:
                filepath = os.path.join(root, filename)
                file_paths.append(filepath)
        return file_paths

    # use: files.get_files_with_extensions({".h", ".c"}) -> ["src/item1.c", "src/headers/item1.h"]
    def get_files_with_extensions(self, extensions):
        files = []
        for file in self.all_files:
            for extension in extensions:
                if file.endswith(extension):
                    files.append(file)
        return files



template = """
import os

from lift.lift_class import LiftClass
import lift.print_color as Out

TOGGLE_DEBUG = True

def setup(lift):
    Out.print_debug(". Running lift_build.py->setup(lift)")
    # Project settings
    lift.app_name = "app"
    lift.dir_root = os.getcwd();
    lift.dir_source = "/src"
    lift.dir_build = "/build"
    lift.dir_lib = ""
    lift.dir_include = ""
    lift.libs = ""

    # Compiler settings
    lift.compiler = "clang"
    lift.flags_debug = lift.flags_default("debug")
    lift.flags_release = lift.flags_default("release")

    # Incremental compilation settings
    lift.incremental_compilation = True

    # Compilation & Linker settings
    lift.threads = 8
    
    # Compiler formatting
    lift.str_format_compiler = "{FLAGS} -c '{C_FILE}' -o '{O_FILE}'"
    lift.str_format_linker = "{DIR_LIB} {LIBS} {DIR_INCLUDE} {FLAGS} {OBJECTS} -o '{APP_NAME}'"

    return lift


def build(lift, mode):
    Out.print_debug(f". Running lift_build.py->build({mode})")
    lift.build(mode)

def run(lift):
    Out.print_debug(". Running lift_build.py->run()")
    args = [""]
    lift.run(args)

def clean(lift):
    Out.print_debug(". Running lift_build.py->clean()")
    lift.clean()

def test(lift):
    Out.print_debug(". Running lift_build.py->test()")
    lift.test()
    
if __name__ == "__main__":
    Out.print_error("This scipt is not designed do be executed with Python itself, its part of the lift build system")
"""



import lift.print_color as Out

import os
import subprocess
import shlex
from threading import Thread
import time

class Workers:
    def internal_worker_capture(self, app, command):
        output, error = self.internal_worker_exe(app, command)
        if output:
            Out.print_normal(output)
        if error:
            Out.print_normal(error)

    def internal_worker_exe(self, app, args_str):
        args = shlex.split(args_str)
        result = subprocess.run([app] + args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        return result.stdout, result.stderr

    def work(self, app, jobs_list, thread_limit):
        threads = []
        Out.print_info(f"> - Total {len(jobs_list)} jobs")
        done = 0
        for job in jobs_list:
            if len(threads) >= thread_limit:
                for t in threads:
                    t.join()
                threads = []
            t = Thread(target=self.internal_worker_capture, args=(app, job))
            threads.append(t)
            t.start()
            done += 1
            Out.print_info(f"> - ↺  {done}/{len(jobs_list)} {int((done / len(jobs_list)) * 100 + 0.5)}%")

        for t in threads:
            t.join()
        Out.print_info("> - All jobs finished")



import os

def get_lift_build_path():
    working_dir = os.getcwd()
    lift_build_path = working_dir + "/lift_build.py"
    return lift_build_path

# TODO

def project_root_dir_path():
    pass

def build_dir_path():
    working_dir = os.getcwd()
    build_dir_path = working_dir + "/build/"
    return build_dir_path


from setuptools import setup, find_packages
 
setup(
    name='lift',
    version='0.0.1',
    author='AutodidactsHaven',
    author_email='author@example.com',
    description='Lift is a CMAKE/makefile replacement for C built with Python (Similar cli to zig build or Rust Cargo)',
    python_requires='>=3.0',
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    url='https://github.com/AutodidactsHaven/lift',
    install_requires="",
    packages=find_packages("src"),
    package_dir={"": "src"},
    entry_points={
        'console_scripts': [
            'lift=lift.lift:main'
        ]
    },
)



import os
import shutil
import lift.print_color as Out
from lift.compiler import Compiler
from lift.files import Files
from lift.workers import Workers
from lift.incremental import Incremental

class LiftClass:
    # Name of your app
    app_name = "app"
    # Root directory of your app (contains "lift_build.py", "build/", "src/")
    dir_root = os.getcwd();
    # Sources directory (usually "/src")
    dir_source = "/src"
    # Build directory (usually "/build")
    dir_build = "/build"
    # Objects directory within build directory (usually "/object") 
    dir_object = "/object"
    # Directory for additonal libs ("-L" flag)
    dir_lib = ""
    # Directory for additonal includes ("-I" flag)
    dir_include = ""
    # List of libs (Example: "-lraylib -lX11")
    libs = ""

    # blacklist files
    file_blacklist = {""}
    # blacklist dirs
    dir_blacklist = {""}

    # Compiler settings
    compiler = "gcc"
    flags_debug = ""
    flags_release = ""
    # Incremental compilation settings
    incremental_compilation = True

    # Compilation & Linker settings
    threads = 8

    # Compiler formatting
    str_format_compiler = "{FLAGS} -c '{C_FILE}' -o '{O_FILE}'"
    str_format_linker = "{DIR_LIB} {LIBS} {DIR_INCLUDE} {FLAGS} {OBJECTS} -o '{APP_NAME}'"


    def __init__(self):
        Out.print_debug("> Generating lift_class.py->__init__")
        # default LiftClass constructor
        self.flags_debug = self.flags_default("debug")
        self.flags_release = self.flags_default("release")
        # TODO: Load preference for gcc/clang from config

    ###########################################################################

    # "release" / "debug"
    def flags_default(self, mode):
        comp = Compiler(self.compiler)
        return comp.generate_flags(mode)

    def build(self, mode):
        Out.print_debug(f"> Running LiftClass.build({mode})")
        # compile object files
        source_files = self.get_source_files()
        files_to_compile = self.get_files_filter(source_files, ".c"); # FIXME: Temporary until Omni is done
        self.compile(mode, files_to_compile)
        # remove old app executable
        self.remove_app()
        # link object files 
        objects_to_link = self.get_object_files()
        self.link(mode, objects_to_link)

    def run(self, args):
        Out.print_debug("> Running LiftClass.run()")
        app_name = self.dir_root + self.dir_build + "/" + self.app_name
        exec_args = [app_name] + args
        # Replace the current process with the executable
        os.execvp(app_name, exec_args)

    def clean(self):
        Out.print_debug("> Running LiftClass.clean()")
        path = self.dir_root + self.dir_build
        if os.path.exists(path):
            # Display the path and ask for confirmation
            response = input(f"{Out.COLOR.RED}> Delete the folder at {Out.COLOR.BLUE}{path}{Out.COLOR.RESET}? (y/n) ")
            if response.lower() == "y":
                # Remove the folder and everything inside it
                shutil.rmtree(path)
                Out.print_info(f"> Deleted folder at {path}")
            else:
                Out.print_info(f"> Cancelled deletion of folder at {path}")
        else:
            Out.print_info(f"> Folder at {path} does not exist")

    def test(self):
        Out.print_debug("> Running LiftClass.test()")

    def compile(self, mode, source_files):
        Out.print_debug(f"> Running LiftClass.compile({mode}, {source_files})")
        self.dir_makedir(self.dir_root + self.dir_build)
        self.dir_makedir(self.dir_root + self.dir_build + self.dir_object)
        # generate compiler command
        commands = []
        for file in source_files:
            o_file = self.dir_root + self.dir_build + self.dir_object + "/" + file.split("/")[-1].replace(".c", ".o")
            command = self.str_format_compiler
            command = command.replace("{FLAGS}", self.get_flags(mode))
            command = command.replace("{C_FILE}", file)
            command = command.replace("{O_FILE}", o_file)
            commands.append(command)
            Out.print_debug(command)
        # spawn workers 
        worker_manager = Workers()
        worker_manager.work(self.compiler, commands, self.threads)
        

    def link(self, mode, object_files):
        Out.print_debug(f"> Running LiftClass.link({mode}, {object_files})")
        self.dir_makedir(self.dir_root + self.dir_build)
        self.dir_makedir(self.dir_root + self.dir_build + self.dir_object)
        # Combine all objects into string
        objects = ""
        for file in object_files:
            objects += f" '{file}' "
        # generate compiler command
        app_name = self.dir_root + self.dir_build + "/" + self.app_name
        command = self.str_format_linker
        command = command.replace("{DIR_LIB}", self.dir_lib)
        command = command.replace("{LIBS}", self.libs)
        command = command.replace("{DIR_INCLUDE}", self.dir_include)
        command = command.replace("{FLAGS}", self.get_flags(mode))
        command = command.replace("{OBJECTS}", objects)
        command = command.replace("{APP_NAME}", app_name)
        Out.print_debug(command)
        # spawn workers 
        worker_manager = Workers()
        worker_manager.work(self.compiler, [command], self.threads)

    def get_source_files(self):        
        # TODO: Excluding elements in (file_blacklist, dir_blacklist) from files
        Out.print_debug("> Running LiftClass.get_source_files()")
        files = Files(self.dir_root + self.dir_source)

        if self.incremental_compilation:
            ### Incremental compilation
            Out.print_info("> Resolving incremental compilation")
            inc = Incremental()
            build_dir = self.dir_root + self.dir_build
            source_files = inc.resolve(files, build_dir)
        else:
            source_files = files.get_files_with_extensions({".c", ".h"})

        return source_files

    def get_object_files(self):
        # TODO: Excluding elements in (file_blacklist, dir_blacklist) from .o files
        # if program was compiled eariles without those filters, these .o files still
        # might exist and become part of the executable by accident.
        Out.print_debug("> Running LiftClass.get_object_files()")
        files = Files(self.dir_root + self.dir_build + self.dir_object)
        object_files = files.get_files_with_extensions({".o"})
        return object_files 

    def get_files_filter(self, files, extension):
        file_list = []
        for file in [f for f in files if f.endswith(extension)]:
            file_list.append(file)
        return file_list

    def get_flags(self, mode):
        if mode == "debug":
            return self.flags_debug
        else:
            return self.flags_release

    # make dir if it does not exist
    def dir_makedir(self, directory):
        if not os.path.exists(directory):
            os.mkdir(directory)

    def remove_app(self):
        app_name = self.dir_root + self.dir_build + "/" + self.app_name
        if os.path.exists(app_name):
            os.remove(app_name)



import os
import re

"""
Given the example layout in /sample we can create an include graph which looks something like this

  ---------         ---------
 | magic.c | ----> | magic.h | <--------|
  ---------         ---------           |
                        ^            --------
                        |           | main.c |
  ----------        ----------       --------
 | player.c | ---> | player.h | <-------|
  ----------        ----------

It is a dependency graph such that we can ascertain *which* c files must be recompiled when a particular .h file has been
modified by traversing the graph.

If a '.c' source file has been modified, then only itself is recompiled
If a '.h' header file has been modified, the graph must be walked to find all header files that include that header file
                                         either directly or via another include. any '.c' source files that depend on any
                                         of those header files must be recompiled
E.g.

If player.c is modified, only player.o must be recompiled
If magic.c is modified, only magic.o must be recompiled
if main.c is modified, only main.o must be recompiled
If player.h is modified, player.o and main.o must be recompiled 
Finally, if magic.h is modified, magic.o, player.o, and main.o all must be recompiled.
This is because player.h #includes magic.h, meaning a change to magic.h will modify the contents of player.h after the preprocessor has run.
Any source file depending on player.h must therefore also be recompiled.
"""

class FileNode:
    def __init__(self, absolute_path, filename):
        self.absolute_path = absolute_path
        self.filename = filename
        
    def __str__(self) -> str:
        return self.absolute_path

    def __repr__(self) -> str:
        return f'{self.filename}'

class DependencyGraph:
    def __init__(self):
        self.num_of_nodes = 0
        self.name_to_node_map = {}
        self.graph = {}

    def add_nodes(self, nodes):
        for node in nodes:
            self.name_to_node_map[node.filename] = node
            self.graph[node] = set()

    """ node2 depends on node1 """
    def add_dependency(self, node1, node2, directional=False):
        self.graph[node2].add(node1)

        if not directional:
            self.graph[node1].add(node2)

    def get_node_by_filename(self, filename):
        if self.name_to_node_map.get(filename):
            return self.name_to_node_map[filename]
        else:
            return None
    
    def get_node_by_absolute_path(self, p):
        for node in self.graph:
            if node.absolute_path == p:
                return node
        return None

    """returns all children .c files that depend on start_node"""
    def accumulate_dependencies(self, start_node):
        visited = {}
        deps = self._accumulate_dependencies(start_node, visited)
        return deps

    def _accumulate_dependencies(self, node, visited):
        deps = []
        for dependent in self.graph[node]:
            # print(f'{node.filename} > {dependent.filename}')
            absolute_path = dependent.absolute_path
            if absolute_path not in visited:
                visited[absolute_path] = True
                child_deps = self._accumulate_dependencies(dependent, visited)
                for d in child_deps:
                    deps.append(d)
            # else already accounted for

            if absolute_path.endswith('.c'):
                deps.append(absolute_path)

        return deps

    def __str__(self) -> str:
        print_string = "Dependency Graph:\n"
        for node in self.graph:
            print_string += f'  {node.absolute_path} has these dependents [{self.graph[node]}]\n'
        return print_string